/*
** Module   : FrameWidget.java
** Abstract : server side frame widget class (contains the configuration)
**
** Copyright (c) 2005-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 NVS 20051026   @23491 Created initial version.
** 002 NVS 20060114   @24097 Frame is created hidden.
** 003 SIY 20060213   @24581 Fixes for the DialogBox support.
** 004 SIY 20060426   @25911 Changed signatures of get/setNextTabItem.
** 005 SIY 20060616   @27537 Added isDown(),setName() methods.
** 006 GES 20060803   @28377 Some cleanup and refactoring.
** 007 ECF 20061207   @31598 Fixed setAttrSpace(). Was calling the wrong method in FrameConfig.
** 008 CA  20090413   @41775 Fixes for FRAME:HIDDEN attribute. On init, HIDDEN flag is off. The
**                           setter must change the HIDDEN flag, not VISIBLE.
** 009 GES 20101005          Added a non-functional COM-HANDLE attribute stub.
** 010 SIY 20101006          Added moveAfterTab() stub.
** 011 SVL 20101006          #10 moved to GenericWidget.
** 012 SIY 20101007          Added setName() with different signature.
** 013 SIY 20111101          Updated imports.
** 014 CA  20130222          Added NPE protection for cases when a widget is used before is 
**                           attached to a frame (after a CREATE WIDGET statement is used).
** 015 CA  20130927          Added LegacyResource annotation, to determine resource type.
** 016 SVL 20131203          Changes caused by renaming of Nameable.get/setName to Nameable.name.
** 017 CA  20131115          Changes related to properly instantiate dynamic resources.
**                           Fixed implicit deletion of static frames.
** 018 CA  20140326          Fixed implicitDeletion in case is called from an Agent, outside of
**                           an external procedure (when the remote persistent procedure is
**                           deleted by the Agent).
** 019 GES 20140627          Inherit from PaneEntity and move all of ContainerEntity into
**                           this class.
** 020 EVL 20140926          More support for dynamic widget cleanup - dynamic frame must be
**                           deregistred in LogicalTerminal.
** 021 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 022 GES 20141124          Added FrameInterface and the centered attr.
** 023 CA  20150312          An unrealized frame needs to be deregistered explicitly when the 
**                           frame gets deleted.
** 024 CA  20150416          Fixed server-side support for PARENT attribute.
**                           Fixed support for NEXT/PREV-SIBLING in relation with tree widget
**                           structures.
** 025 HC  20150509          Javadoc fix.
** 026 GES 20150506          Comment and javadoc cleanup.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;

/**
 * This is the server side frame widget class which contains the 
 * client-side configuration for this frame BUT which does NOT implement
 * any of the server side logic for frame processing.
 * <p>
 * WARNING: unlike most of the server-side widget classes, this class
 * intentionally DOES NOT push a modified screen definition to the client
 * when visibility/hidden/down changes occur.  Such changes transfer to the
 * client using alternate paths and since these are frequently changed, using
 * the screen definition for such things is not acceptable.
 */
@LegacyResource(resource = LegacyResource.FRAME)
public class FrameWidget
extends PaneEntity<FrameConfig>
implements FrameInterface,
           CommonHandleTree
{
   /**
    * Default constructor which can be used to create a static frame.
    */
   public FrameWidget()
   {
      this(false);
   }
   
   /**
    * Primary constructor.
    * 
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    */
   public FrameWidget(boolean dynamic)
   {
      super(dynamic, new FrameConfig());
      config.hidden = false;
   }
   
   /**
    * Set NAME attribute.
    * 
    * @param   name
    *          New value for the attribute.
    */
   public void name(String name)
   {
      config.name = name;
   }
   
   /**
    * Set NAME attribute.
    * 
    * @param   name
    *          New value for the attribute.
    */
   public void name(character name)
   {
      name(name.isUnknown() ? "" : name.toStringMessage());
   }
   
   /**
    * Gets the NEXT-TAB-ITEM writable attribute.
    *
    * @return   The current value of NEXT-TAB-ITEM attribute.
    */
   public GenericWidget<?> getNextTabItem()
   {
      // TODO: implement this!
      throw new RuntimeException("NEXT-TAB-ITEM is not done yet!");
   }

   /**
    * Sets the NEXT-TAB-ITEM writable attribute.
    *
    * @param    n
    *           The new value for the NEXT-TAB-ITEM attribute.
    */
   public void setNextTabItem(GenericWidget<?> n)
   {
      // TODO: implement this!
      throw new RuntimeException("NEXT-TAB-ITEM is not done yet!");
   }

   /**
    * Gets the ATTR-SPACE writable attribute.
    *
    * @return   The current value of the ATTR-SPACE attribute.
    */
   public logical isAttrSpace()
   {
      return new logical(config.attr_space);
   }

   /**
    * Sets the ATTR-SPACE writable attribute.
    *
    * @param    space
    *           The new value for the ATTR-SPACE attribute.
    */
   public void setAttrSpace(logical space)
   {
      setAttrSpace(space.booleanValue());
   }

   /**
    * Sets the ATTR-SPACE writable attribute.
    *
    * @param    space
    *           The new value for the ATTR-SPACE attribute.
    */
   public void setAttrSpace(boolean space)
   {
      config.attr_space = space;
      pushScreenDefinition();
   }
   
   /**
    * Get the value of the CENTERED attribute which defines if the frame is to be centered in
    * the width of the parent window.
    *
    * @return   <code>true</code> if the frame is centered.
    */
   public logical isCentered()
   {
      return new logical(config.centered);
   }

   /**
    * Set the value of the CENTERED attribute which defines if the frame is to be centered in
    * the width of the parent window.
    *
    * @param    center
    *           <code>true</code> if the frame is to be centered.
    */
   public void setCentered(boolean center)
   {
      if (config.centered && !center)
      {
         UnimplementedFeature.todo("4GL docs report this should gen an error when setting " +
                                   "from true to false.");
      }
      
      config.centered = center;
      pushScreenDefinition();
   }

   /**
    * Set the value of the CENTERED attribute which defines if the frame is to be centered in
    * the width of the parent window.
    *
    * @param    center
    *           <code>true</code> if the frame is to be centered.
    */
   public void setCentered(logical center)
   {
      if (center == null || center.isUnknown())
      {
         UnimplementedFeature.todo("CENTERED attribute unknown behavior is not known.");
         return;
      }
      
      setCentered(center.booleanValue()); 
   }
   
   /**
    * Sets the dialog box flag.
    *
    * @param dialogBox
    *        new value for the dialog box flag
    */
   public void setDialogBox(boolean dialogBox)
   {
      config.dialog = dialogBox;
   }
   
   /**
    * Gets the dialog box flag.
    *
    * @return <code>true</code> if this frame is a dialog box.
    */
   public boolean isDialogBox()
   {
      return config.dialog;
   }

   /**
    * Sets the down frame size.
    *
    * @param    down
    *           The new number of iterations for this down frame.
    */
   public void setDown(int down)
   {
      config.down = down;
   }
   
   /**
    * Gets the down frame size.
    *
    * @return   The number of data iterations for the frame (1 if the
    *           frame is not a down frame).
    */
   public int getDown()
   {
      return config.down;
   }

   /**
    * Return <code>true</code> if frame is DOWN frame.
    * 
    * @return   <code>true</code> if frame is DOWN frame.
    */
   public boolean isDown()
   {
      return (getDown() != 1);
   }
   
   /**
    * Sets the new value for HIDDEN flag.
    * 
    * @param   hidden
    *          The new boolean value for the HIDDEN attribute.
    */
   public void setHidden(boolean hidden)
   {
      config.hidden = hidden;
   }

   /**
    * Sets the Progress VISIBLE attribute.
    * 
    * @param   visible
    *          The new value for the Progress VISIBLE attribute.
    */
   public void setVisible(boolean visible)
   {
      config.visible = visible;
   }
   
   /**
    * Gets the Progress COM-HANDLE attribute.
    * <p>
    * This is a non-functional stub for now.
    * 
    * @return   The value for the COM-HANDLE attribute.
    */
   public comhandle getComHandle()
   {
      return new comhandle();
   }
   
   /**
    * Hides this widget.
    * <p>
    * This method corresponds to the Progress statement: HIDE widget
    * 
    * @param   noPause
    *          <code>true</code> causes omission of any pauses which would
    *          be necessary otherwise.
    */
   @Override
   public void hide(boolean noPause)
   {
      frame.hide(noPause);
   }

   /**
    * Gets the SELECTED writable attribute.
    *
    * @return    The current value of the SELECTED attribute.
    */
   @Override
   public logical isSelected()
   {
      UnimplementedFeature.partial("SELECTED attribute client-side runtime is not implemented");
      return new logical(config.selected);
   }

   /**
    * Sets the SELECTED writable attribute.
    *
    * @param    sel
    *           The new value for the SELECTED attribute.
    */
   @Override
   public void setSelected(logical sel)
   {
      if (sel == null || sel.isUnknown())
      {
         UnimplementedFeature.todo("SELECTED attribute unknown behavior is not known.");
         return;
      }

      setSelected(sel.booleanValue());
   }

   /**
    * Sets the SELECTED writable attribute.
    *
    * @param    sel
    *           The new value for the SELECTED attribute.
    */
   @Override
   public void setSelected(boolean sel)
   {
      UnimplementedFeature.partial("SELECTED attribute client-side runtime is not implemented");
      
      config.selected = sel;
      pushScreenDefinition();
   }
   
   /**
    * Queries if the user is allowed to move this widget using direct manipulation.
    * 
    * @return  If <code>true</code>, the widget is movable via direct manipulation.
    */
   public logical isMovable()
   {
      UnimplementedFeature.partial("MOVABLE attribute client-side runtime is not implemented");
      return new logical(config.movable);
   }
   
   /**
    * Defines if the user is allowed to move this widget using direct manipulation.
    * 
    * @param   movable
    *          <code>true</code> to enable movable direct manipulation.
    */
   public void setMovable(boolean movable)
   {
      UnimplementedFeature.partial("MOVABLE attribute client-side runtime is not implemented");
      
      config.movable = movable;
      pushScreenDefinition();
   }
   
   /**
    * Defines if the user is allowed to move this widget using direct manipulation.
    * 
    * @param   movable
    *          <code>true</code> to enable movable direct manipulation.
    */
   public void setMovable(logical movable)
   {
      if (movable == null || movable.isUnknown())
      {
         UnimplementedFeature.todo("MOVABLE attribute unknown behavior is not known.");
         return;
      }
      
      setMovable(movable.booleanValue()); 
   }
   
   /**
    * Gets the SCROLLABLE writeable attribute.
    *
    * @return   The current value of the SCROLLABLE attribute.
    */
   public logical isScrollable()
   {
      return new logical(config.scrollable);
   }

   /**
    * Sets the SCROLLABLE writeable attribute.
    *
    * @param    scroll
    *           The new value for the SCROLLABLE attribute.
    */
   public void setScrollable(logical scroll)
   {
      // TODO: this will not handle unknown value well
      setScrollable(scroll.booleanValue());
   }

   /**
    * Sets the SCROLLABLE writeable attribute.
    *
    * @param    scroll
    *           The new value for the SCROLLABLE attribute.
    */
   public void setScrollable(boolean scroll)
   {
      config.scrollable = scroll;
      pushScreenDefinition();
   }
   
   /**
    * Accesses the first widget in the list of contained widgets.
    *
    * @return   The first widget or <code>null</code> if no widgets are contained.
    */
   @Override
   public handle firstChild()
   {
      return HandleChain.firstResource(this, LegacyResource.FRAME);
   }
   
   /**
    * Accesses the last widget in the list of contained widgets.
    *
    * @return   The last widget or <code>null</code> if no widgets are contained.
    */
   @Override
   public handle lastChild()
   {
      return HandleChain.lastResource(this, LegacyResource.FRAME);
   }
   
   /**
    * Sets the containing frame instance.
    * 
    * @param    frame
    *           The containing instance.
    */
   @Override
   public void setFrame(GenericFrame frame)
   {
      super.setFrame(frame);
      
      if (frame.asWidget() != this)
      {
         this.config.parentId = frame.firstChild().getId();
      }
   }
   
   /**
    * Indicates an implementation of siblings methods.
    * 
    * @return   Always <code>true</code>. 
    */
   @Override
   protected boolean useHandleChainSiblings()
   {
      return true;
   }

   /**
    * Check if the resource may be implicitly deleted. 
    * 
    * @return   <code>true</code> if the INSTANTIATING-PROCEDURE attribute is unknown (i.e. the 
    *           instantiating procedure has been deleted).
    */
   protected boolean implicitDeletion()
   {
      // shared frames can be declared only in external procedure; they are always marked as for
      // implicit deletion; resourceDelete will take care to delete only the first reference
      return frame.isShared()                          || 
             TransactionManager.getNestingLevel() == 0 || 
             !TransactionManager.isExternalBlock()     || 
             instantiatingProcedure().isUnknown();
   }

   /**
    * Delete the resource.
    * 
    * @return   <code>true</code> if the resource was deleted.
    */
   @Override
   protected boolean resourceDelete()
   {
      if (!canDelete())
      {
         return false;
      }
      
      // check frame special cases
      if (!_dynamic()        && 
          _isVisible()       &&  
          !frame.isShared()  &&
          !(frame.isRedirected() || frame.isOutRedirected()))
      {
         // implicit static frame deletion is possible only if is invisible
         return false;
      }
      
      if (frame.isShared() && frame.getScopeNesting() > 0)
      {
         // shared frames can't be deleted until all are out of scope.
         return false;
      }
      
      deleted = true;
      
      // remove the frame, etc
      frame.frameCleanup();
      
      if (_dynamic() || !config.realized)
      {
         LogicalTerminal.deregisterFrame(getId());
      }

      ConfigManager.getInstance().removeWidgetConfig(config.id);
      
      return true;
   }

   /**
    * Set the instantiating procedure for this resource.
    * <p>
    * This will set the instantiating procedure only if is not shared or this is the procedure
    * where the frame is registered as shared.
    *  
    * @param    hproc
    *           The procedure where the resource was instantiated.
    */
   @Override
   protected void setInstantiatingProcedure(handle hproc)
   {
      if (_dynamic() || !frame.isShared() || frame.getScopeNesting() == 0)
      {
         this.instProc.assign(hproc);
      }
   }

   /**
    * Force new instantiating procedure for this frame.
    * 
    * @param    hproc
    *           The procedure where the resource was instantiated. For shared frames, this must be
    *           the procedure where the frame was registered as shared.
    */
   protected void setSharedInstantiatingProcedure(handle hproc)
   {
      this.instProc.assign(hproc);
   }

   /**
    * Check if the specified widget can act as parent for this widget.
    *
    * @param    widget
    *           The new value for the PARENT attribute.
    * 
    * @return   <code>true</code> if the widget is a FIELD-GROUP or WINDOW.
    */
   @Override
   protected boolean validParent(GenericWidget<?> widget)
   {
      // parent must be a field-group or window
      String ptype = widget.type();
      return ptype.equals(LegacyResource.FIELD_GROUP) || ptype.equals(LegacyResource.WINDOW);
   }
}
