/*
** Module   : AbstractChuiDriver.java
** Abstract : simple base class for ChUI screen driver implementation
**
** Copyright (c) 2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 GES 20150402 Created initial version.
*/

package com.goldencode.p2j.ui.client.chui.driver;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.MouseEvent;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.chui.*;

/**
 * Simple base class for ChUI-specific screen driver implementation features. GUI-specific
 * features are not to be implemented here.  This class is intended to be specific to a given
 * UI-modality (ChUI in this case). However, no driver-specific implementation classes or features
 * should be present.  For example, no Swing-specific feature usage is allowed.
 */
public abstract class AbstractChuiDriver
extends AbstractDriver<BasePrimitives, ChuiOutputManager>
{
   /**
    * Construct a new instance.
    *
    * @param    direct
    *           The output primitives.
    * @param    term
    *           Terminal type string.
    */
   public AbstractChuiDriver(BasePrimitives direct, String term)
   {
      super(new ChuiWidgetFactory(), direct, true, term);
   }

   /**
    * Read configuration values from the bootstrap configuration and convert
    * these into an options instance.
    * <p>
    * The following parameters are supported:
    * <p>
    * <pre>
    * Category:Group:Key       Type
    * -----------------------  ---------------------------
    * client:chui:rows         integer
    * client:chui:columns      integer
    * client:chui:background   text which can be converted to a ColorRgb by ColorRgb.fromString()
    * client:chui:foreground   text which can be converted to a ColorRgb by ColorRgb.fromString()
    * client:chui:selection    text which can be converted to a ColorRgb by ColorRgb.fromString()
    * client:chui:fontname     text containing a monospaced (fixed width) font family name
    * client:chui:fontsize     integer
    * </pre>
    *
    * @param    cfg
    *           The bootstrap configuration.
    *
    * @return   A new options instance with any defaults overridden as specified
    *           in the bootstrap configuration.
    */
   public static TerminalOptions convertToOptions(BootstrapConfig cfg)
   {
      TerminalOptions options = new TerminalOptions();
      
      String next;
      
      if ((next = cfg.getString("client", "chui", "rows", null)) != null)
      {
         options.rows = parseInt(next, options.rows);
      }
      
      if ((next = cfg.getString("client", "chui", "columns", null)) != null)
      {
         options.columns = parseInt(next, options.columns);
      }
      
      if ((next = cfg.getString("client", "chui", "background", null)) != null)
      {
         options.background = ColorRgb.fromString(next, options.background);
      }
      
      if ((next = cfg.getString("client", "chui", "foreground", null)) != null)
      {
         options.foreground = ColorRgb.fromString(next, options.foreground);
      }
      
      if ((next = cfg.getString("client", "chui", "selection", null)) != null)
      {
         options.selection = ColorRgb.fromString(next, options.selection);
      }
      
      if ((next = cfg.getString("client", "chui", "fontname", null)) != null)
      {
         options.fontname = next;
      }
      
      if ((next = cfg.getString("client", "chui", "fontsize", null)) != null)
      {
         options.fontsize = parseInt(next, options.fontsize);
      }
      
      if ((next = cfg.getString("client", "chui", "windows", null)) != null)
      {
         options.underWindows = "true".equalsIgnoreCase(next);
      }
      
      return options;
   }

   /**
    * Get the {@link OutputManager} instance associated with this driver.
    * 
    * @param    cfg
    *           The initial configuration for this driver.
    *
    * @return   The instance.
    */
   @Override
   public ChuiOutputManager createOutputManager(BootstrapConfig cfg)
   {
      return new ChuiOutputManager(this);
   }
   
   /**
    * Set cursor status (ON/OFF).
    * <p>
    * For GUI, this does nothing.
    * 
    * @param   on
    *          <code>true</code> - show cursor, <code>false</code> - hide cursor.
    *
    * @return  Always <code>false</code>.
    */
   @Override
   public boolean setCursorStatus(boolean on)
   {
      return false;
   }

   /**
    * Enable or disable capture of mouse events.
    * <p>
    * This implementation is a no-op.
    *  
    * @param    state
    *           <code>true</code> to capture mouse events, <code>false</code> to ignore them.
    *           
    * @return   The previous state.
    */
   @Override
   public boolean captureMouseEvents(boolean state)
   {
      // no-op
      return false;
   }

   /**
    * Handle the specified event, in the context of the specified widget.
    * <p>
    * This implementation is a no-op.
    * 
    * @param    sourceID
    *           The source widget ID.
    * @param    evt
    *           The mouse event.
    */
   @Override
   public void handleMouseEvent(int sourceID, MouseEvent evt)
   {
      // no-op
   }
   
   /** 
    * Get the native screen size, in native units.
    * 
    * @return   See above.
    */
   @Override
   public NativeDimension getScreenSize()
   {
      ScreenBitmap bitmap = direct.getScreenBitmap();
      
      return new NativeDimension(bitmap.cols(), bitmap.rows());
   }
   
   /**
    * Attempt to convert the given string into an integer using
    * <code>Integer.parseInt</code>.
    *
    * @param    spec
    *           String encoded number specification.
    * @param    def
    *           Default value to return if the spec cannot be parsed properly.
    *
    * @return   The parsed number or the default if there is any parsing
    *           problem.
    */
   private static int parseInt(String spec, int def)
   {
      int val = def;
      
      try
      {
         val = Integer.parseInt(spec);
      }
      
      catch (NumberFormatException nfe)
      {
         // ignore
      }
      
      return val;
   }
}
