/*
** Module   : BatchDriver.java
** Abstract : non-interactive (batch mode) driver
**
** Copyright (c) 2013-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20131104 Created initial version.
** 002 EVL 20140114 The key reader should return -2(EOF) in regular batch mode. Also we need to
**                  separate processing for background and foreground mode simulation. The native
**                  library loading has been moved to central location.
**                  See the ClientCore class.
** 003 EVL 20140224 The batch driver does not need to initialize console in Windows.  The Windows
**                  console modifies the screen in a way not acceptable for batch driver. Also we
**                  do not need to clear the screen in batch mode.  So the initChui() nas been
**                  removed.  The new method terminalType() returning empty string has been added
**                  to batch driver.  Both Linux and Windows 4GL return empty string for TERMINAL
**                  function.  Also in the Windows foreground mode we need to have the stub for
**                  setCursorStatus() method.
** 004 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
**                  Removed hasColors, as it is not used.
** 005 CA  20141029 Removed the redundant ScreenDriver interface implementation.
** 006 GES 20150331 Moved initialization logic here from OutputManager to separate concerns.
**                  Reworked to use a common base class for more shared code.
*/

package com.goldencode.p2j.ui.client.chui.driver.batch;

import com.goldencode.p2j.ui.client.chui.driver.console.*;
import com.goldencode.p2j.cfg.*;
import com.goldencode.util.*;

/**
 * The dummy console pseudo batch driver.  The underlying implementation is used to provide I/O
 * handling for batch mode.
 * <p>
 * During initialization, the Progress 4GL client in batch mode outputs a Progress logo,
 * copyright notice and version when run in the foreground (and just the version when run in the
 * background).  This implementation is deliberately NOT duplicating these outputs for these
 * reasons:
 * <ol>
 *   <li> The output has no value for application/user processing or functionality.
 *   <li> Generating that output would confuse people as to who owns this technology.
 *   <li> The output content changes from version to version anyway, so it is very unlikely that
 *        an application would depend upon the output.      
 * </ol>
 */
public class BatchDriver
extends ConsoleDriver
{
   /**
    * Constructor. 
    * 
    * @param   cfg
    *          The <code>BootstrapConfig</code> configuration.
    */
   public BatchDriver(BootstrapConfig cfg)
   {
      super(new BatchPrimitives(), "");
      
      // always in batch mode
      setBatchMode(true);
      
      // mark batch session as running in background or not
      setBackgroundMode(cfg.getBoolean("client", "driver", "background", false));
   }

   /**
    * Generate a beep signal. 
    * <p>
    * Implementation deoends on native layer and (in turn) from particular 
    * terminal. If terminal does not support this feature or sound in terminal
    * is turned off, there will be no audible feedback to user.
    */
   @Override
   public void beep()
   {
      BatchHelper.outputOperation();
   }

   /**
    * Clear screen.
    */
   @Override
   public void clear()
   {
      if (direct != null)
      {
         if (((BatchPrimitives) direct).isActive())
         {
            BatchHelper.outputOperation();
         }
      }
      else
      {
         super.clear();
      }
   }

   /**
    * Read the key from terminal and translate it into P2J internal 
    * representation. Translation details are in {@link KeyProcessor} class. 
    * Mouse events (which also returned from native layer) are filtered out 
    * because mouse is not supported.  
    */
   @Override
   public int readKey()
   {
      if (direct != null)
      {
         if (((BatchPrimitives) direct).isActive())
         {
             // we are under Windows or Linux foreground mode
            if (PlatformHelper.isUnderWindowsFamily() || !inBackgroundMode())
            {
               ((BatchPrimitives) direct).deActivate();
               
               // in Windows we have to generate the error and stop client.
               BatchHelper.inputOperation();
            }
            return -2;
         }
         else
         {
            // otherwise pass control to ConsoleDriver key handler 
            return super.readKey();
         }
      }
      else
      {
         return super.readKey();
      }
   }

   /**
    * Reset screen mode.
    */
   @Override
   public void resetMode()
   {
      BatchHelper.outputOperation();
   }

   /**
    * Resume terminal after {@link #suspend()} call.
    */
   @Override
   public void resume()
   {
      if (direct != null)
      {
         if (((BatchPrimitives) direct).isActive())
         {
            BatchHelper.outputOperation();
         }
      }
      else
      {
         super.resume();
      }
   }

   /**
    * Suspend terminal and release it for use by other applications.
    */
   @Override
   public void suspend()
   {
      if (direct != null)
      {
         if (((BatchPrimitives) direct).isActive())
         {
            BatchHelper.outputOperation();
         }
      }
      else
      {
         super.suspend();
      }
   }

   /**
    * Set cursor status (ON/OFF).
    */
   @Override
   public boolean setCursorStatus(boolean on)
   throws IllegalStateException
   {
      return true;
   }
}
