/*
** Module   : ConsoleDriver.java
** Abstract : native console driver (uses JNI helpers to OS console APIs)
**
** Copyright (c) 2010-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20130723 Created initial version
** 002 CA  20131004 Implement the WINDOW-SYSTEM attribute for this driver.
** 003 EVL 20131212 Added method to check if the driver is in batch mode.
** 004 EVL 20140115 Added method to check if the driver is in background mode.
** 005 HC  20140119 Implemented runtime support for the COLOR 
**                  protermcap-attribute.
** 006 MAG 20140512 Add method to check screen interface type ChUI or GUI. 
** 007 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
**                  Removed hasColors, as it's not used; the ChUI drivers rely on 
**                  ConsoleHelper.hasColors. 
** 008 CA  20141123 Added API to retrieve the available screen size where windows can be created.
**                  In ChUI, this is the default window's size.
** 009 GES 20150506 Reworked to use a common base class for more shared code.
*/

package com.goldencode.p2j.ui.client.chui.driver.console;

import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.util.*;

/**
 * Text console-based driver. The underlying native implementation is used as backend to provide
 * required features. The example is NCURSES package for Linux or console API wor Windows OS.
 */
public class ConsoleDriver
extends AbstractChuiDriver
{
   /** Special key code which denotes mouse event. */
   private static final int KEY_MOUSE = 0631;

   /** Key processor. */
   private KeyProcessor processor = new KeyProcessor();

   /** Reference to console low-level driver. */
   private ConsoleHelper helper = ConsoleHelper.getDefaultConsoleHelper();

   /**
    * Constructor. 
    */
   public ConsoleDriver()
   {
      super(new ConsolePrimitives(), null);
   }

   /**
    * Construct a new instance.
    *
    * @param    direct
    *           The output primitives.
    * @param    term
    *           Terminal type string.
    */
   public ConsoleDriver(ConsolePrimitives direct, String term)
   {
      super(direct, term);
   }

   /**
    * Generate a beep signal. 
    * <p>
    * Implementation deoends on native layer and (in turn) from particular 
    * terminal. If terminal does not support this feature or sound in terminal
    * is turned off, there will be no audible feedback to user.
    */
   @Override
   public void beep()
   {
      helper.beep();
   }

   /**
    * Clear screen.
    */
   @Override
   public void clear()
   {
      helper.clear();
   }

   /**
    * Get driver-specific child process factory.
    * 
    * @return   The reference to the factory.
    */
   @Override
   public ChildProcessFactory getChildProcessFactory()
   {
      return new ChildProcessFactory()
      {
         /**
          * Obtains a new instance of an interactive child process.
          *
          * @return   The new child process instance.
          */
         public InteractiveChildProcess getChildProcess()
         {
            return new NativeChildProcess(ConsoleDriver.this);
         }
      };
   }
   
   /**
    * Read the key from terminal and translate it into P2J internal 
    * representation. Translation details are in {@link KeyProcessor} class. 
    * Mouse events (which also returned from native layer) are filtered out 
    * because mouse is not supported.  
    */
   @Override
   public int readKey()
   {
      int key = helper.readKey();

      if (key == ConsoleDriver.KEY_MOUSE || key == -1)
         return -1;

      return processor.map(key);
   }

   /**
    * Reset screen mode.
    */
   @Override
   public void resetMode()
   {
      helper.resetMode();
   }

   /**
    * Resume terminal after {@link #suspend()} call.
    */
   @Override
   public void resume()
   {
      helper.resume();
   }

   /**
    * Set cursor status (ON/OFF).
    */
   @Override
   public boolean setCursorStatus(boolean on)
   throws IllegalStateException
   {
      return helper.setCursorStatus(on);
   }

   /**
    * Set terminal type according to passed string. 
    */
   @Override
   public void setTerminalType(String name)
   {
      ConsoleHelper.setTermType(name);
   }

   /**
    * Suspend terminal and release it for use by other applications.
    */
   @Override                             
   public void suspend()
   {
      helper.suspend();
   }

   /**
    * Get terminal type string.
    */
   @Override
   public String terminalType()
   {
      return ConsoleHelper.getTermType();
   }

   /**
    * Initialize driver.
    */
   @Override
   public void init()
   {
      ConsoleHelper.initChui();
      ((ConsolePrimitives) direct).init(helper);

      if ("vt220".equalsIgnoreCase(terminalType()))
         processor.setDelWorkaroundActive(true);
   }
}
