/*
** Module   : EmbeddedWebServerImpl.java
** Abstract : embedded web server shared implementation
**
** Copyright (c) 2013-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 GES 20150312 Created initial version by moving code from the ChUI web driver and making it
**                  agnostic to web client type.
*/

package com.goldencode.p2j.ui.client.driver.web;

// selective import to avoid name conflicts
import java.util.logging.Logger;
import java.util.logging.Level;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.web.*;

import org.eclipse.jetty.util.resource.*;
import org.eclipse.jetty.server.handler.*;

/**
 * Web screen driver shared implementation of an embedded web server. 
 */
public class EmbeddedWebServerImpl
implements EmbeddedWebServer
{
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(EmbeddedWebServerImpl.class.getName());
   
   /** URI format specification. */
   private static final String spec = "https://%s:%d%s";
   
   /** WebSocket target. */
   private static final String AJAX_TARGET = "/ajax";
   
   /** Embedded web server instance. */
   private GenericWebServer server = new GenericWebServer();
   
   /** Driver name to use in logging. */
   private String name = null;
   
   /** WebSocket factory. */
   private GenericWebSocketCreator sfact = null;
   
   /** Classpath resource instance. */
   private Resource cpres = null;
   
   /** Page handlers for our specific targets. */
   private WebPageHandler[] hdlrs = null;
   
   /**
    * Create an instance.
    *
    * @param    name
    *           Driver name to use in logging.
    * @param    sfact
    *           WebSocket factory.
    * @param    cpres
    *           Classpath resource instance.
    * @param    hdlrs
    *           Page handlers for our specific targets.
    */
   public EmbeddedWebServerImpl(String                  name,
                                GenericWebSocketCreator sfact,
                                Resource                cpres,
                                WebPageHandler[]        hdlrs)
   {
      this.name  = name;
      this.sfact = sfact;
      this.cpres = cpres;
      this.hdlrs = hdlrs;
   }
   
   /**
    * Startup the embedded web server using the application server's <code>KeyStore</code> to
    * build a suitable SSL context.
    * 
    * @param    keyStore
    *           The exported server KeyStore. 
    * @param    config
    *           Client configuration.
    * @param    port
    *           An explicit port to start the web server or 0 to let the OS assign one.
    * @param    host
    *           An explicit host to start the web server or <code>null</code> to use the server's
    *           host.
    */
   @Override
   public String startupServer(ServerKeyStore keyStore, BootstrapConfig config, String host, int port)
   {
      String uri = null;
      
      // startup embedded server
      try
      {
         // authorization token
         String token = keyStore.getAuthorizationToken();
         
         // configure and startup server
         server.setKeyStore(keyStore);
         server.setBootstrapConfig(config);
         
         // process each handler
         for (int i = 0; i < hdlrs.length; i++)
         {
            hdlrs[i].setAuthorizationToken(token);
            
            ContextHandler contextHandler = new ContextHandler();
            contextHandler.setHandler(hdlrs[i]);
            
            server.addHandler(contextHandler);
         }
         
         server.addHandler(new DojoToolkitHandler());
         
         ResourceHandler cliRes = new ResourceHandler();
         cliRes.setDirectoriesListed(false);         
         
         ContextHandler cliCtx = new ContextHandler();
         cliCtx.setContextPath("/client");
         cliCtx.setBaseResource(cpres);
         cliCtx.setHandler(cliRes);
         
         ResourceHandler commonRes = new ResourceHandler();
         commonRes.setDirectoriesListed(false);         
         
         ContextHandler commonCtx = new ContextHandler();
         commonCtx.setContextPath("/common");
         commonCtx.setBaseResource(calcResourceBase(EmbeddedWebServerImpl.class, "res"));
         commonCtx.setHandler(commonRes);
         
         ContextHandlerCollection list = new ContextHandlerCollection();
         list.setHandlers(new ContextHandler[] { cliCtx, commonCtx });
         
         server.addHandler(list);
         
         server.addWebSocketHandler(AJAX_TARGET, sfact);
         
         // TODO: should we add a DefaultHandler instance at the end to get 404 processing?
         
         server.startup(host, port);
      }
      
      catch (Exception e)
      {
         LOG.logp(Level.SEVERE,
                  name,
                  "startupServer",
                  LogHelper.generate("Cannot startup server."),
                  e);         
      }
      
      finally
      {
         uri = this.getServerURI();
      }
      
      return uri;
   }

   /**
    * Shutdown the embedded web server.
    */
   @Override
   public void shutdownServer()
   {
      try
      {
         server.shutdown();
      }
      catch (Exception e)
      {
         LOG.logp(Level.SEVERE,
                  name,
                  "shutdownServer",
                  LogHelper.generate("Cannot shutdown server."),
                  e);         
      }
   }

   /**
    * Get the embedded web server's URI.
    * 
    * @return   A string representing the server's URI.
    */
   @Override
   public String getServerURI()
   {
      String uri = null;
      
      if (server.isRunning())
      {
         uri = String.format(spec, server.getHost(), server.getPort(), server.getContextPath());
      }
      
      return uri;
   }
   
   /**
    * Calculate the path in the jar file for the given path as an extention of the origin
    * class' package and return that resource base as a classpath resource. The result can
    * be used in a resource handler as the location for accessing static resources.
    *
    * @param    origin
    *           A class in the package being targeted as the starting path for the resource.
    *           The package will be looked up from this class, any class in that package will
    *           work as the class is not used for anything else.
    * @param    path
    *           The relative sub-directory to add to the package to get the full base path.
    *           This should NOT have either a leading or trailing path separator ('/') included.
    *
    * @return   The calculated resource base.
    */
   public static Resource calcResourceBase(Class<?> origin, String path)
   {
      String pkg = origin.getPackage().getName();
      
      StringBuilder sb = new StringBuilder();
      sb.append(pkg.replace('.', '/')).append('/').append(path).toString();
      
      return Resource.newClassPathResource(sb.toString());
   }
}
