/*
** Module   : GuiOutputManager.java
** Abstract : Provides a generic interface to low level GUI primitives.
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 CA  20140805 Created initial version.  ThinClient is imported explicitly.
** 002 VIG 20140826 Changed default window ID constant to dynamic ID definition.
** 003 MAG 20140912 Implement GUI status bar.
** 004 HC  20140911 GUI coordinate support - data types refactoring.
** 005 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 006 MAG 20140925 Check if window has renderer when process default window.
** 007 CA  20141021 Removed the repaint(Widget) API: all drawing needs to be done via the common
**                  event processing.   Added translatedOrigin, to inform the caller about the 
**                  real location of the current origin point.
** 008 HC  20141021 Implemented window move/position support.
** 009 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 010 CA  20141123 Added API to check if a certain key can raise a STOP condition.
** 011 CA  20141125 Added API to conditionally enable/disable OS-level events.
** 012 EVL 20150107 Fix for dynamic windows renderer missing exception.
** 013 HC  20150323 WidgetId.getId() renamed to asInt().
** 014 GES 20150425 Match refactoring changes elsewhere.
*/

package com.goldencode.p2j.ui.client.gui.driver;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Provides a generic interface to low level graphic user interface primitives.
 */
public class GuiOutputManager
extends OutputManager<GuiPrimitives>
{
   /** Widget registry. */
   private WidgetRegistry<GuiOutputManager> registry = new WidgetRegistry<GuiOutputManager>();

   /** The initial configuration of this manager. */
   private final BootstrapConfig cfg;
   
   /**
    * Create a new output manager instance, with the specified driver.
    *
    * @param    driver
    *           Implements low level input/output primitives.  Must NOT be <code>null</code>.
    * @param    cfg
    *           The initial configuration.
    */
   public GuiOutputManager(GuiDriver driver, BootstrapConfig cfg)
   {
      super(driver);
      
      this.cfg = cfg;
   }
   
   /**
    * Get the context-specific instance, if it exists. In a single-client
    * JVM environment, this is equivalent to a singleton.
    * 
    * @return   The instance or <code>null</code> if none exists.
    */
   public static synchronized GuiOutputManager instance()
   {
      return (GuiOutputManager) defaultTerminal.get();
   }

   /**
    * Process the default window, depending on the output manager type.
    * 
    * @param    window
    *           A window.
    */
   @Override
   public void activateWindow(Window<?> window)
   {
      // TODO: why not simply call window.getId()?
      int windowId = UiUtils.lookupWidgetId(window).asInt();
  
      // this part of code need to be called for dynamically created windows not yet active to
      // register simulator
      realizeWindow(windowId);
      // activate
      ((GuiDriver) this.driver).selectWindow(windowId);
   }

   /**
    * Realize new window, depending on the output manager type.  This is not default window but
    * one is created dynamically.
    * 
    * @param    windowId
    *           A window ID to be realized.
    */
   @Override
   public void realizeWindow(int windowId)
   {
      GuiDriver driver = (GuiDriver) this.driver;
      
      if (!driver.hasWindowEmulator(windowId))
      {
         driver.registerWindow(windowId);
      }
   }

   /**
    * The real location where the origin was translated.  It represents the top-left corner of
    * the last area set by {@link #setDrawingArea}.
    * 
    * @return   See above.
    */
   @Override
   public Point translatedOrigin()
   {
      NativePoint norig = ((GuiDriver) driver).translatedOrigin();
      
      return coordinates().pointFromNative(norig);
   }
   
   /**
    * Limit the available drawing area to the specified boundaries.
    * 
    * @param    bounds
    *           The bounds for the drawing area.
    */
   public void setDrawingArea(Rectangle bounds)
   {
      NativeRectangle nbounds = bounds == null ? null : coordinates().rectangleToNative(bounds);
      ((GuiDriver) driver).setDrawingArea(nbounds);
   }
   
   /**
    * Clear area occupied by specified widget.
    * 
    * @param   widget
    *          Reference widget. 
    */
   @Override
   public void clear(Widget<?> widget)
   {
      if (widget == null || isRedirected())
         return;

      // TODO: which color?
      int windowId = widget.window().config().id.asInt();
      
      clear(windowId, widget.physicalLocation(), widget.physicalDimension(), null);
   }

   /**
    * Access driver-specific widget factory.
    * 
    * @return  reference to widget factory.
    */
   @Override
   public GuiWidgetFactory getFactory()
   {
      return (GuiWidgetFactory) driver.getFactory();
   }

   /**
    * Get the widget registry for this output manager.
    * 
    * @return   See above.
    */
   @Override
   public WidgetRegistry<GuiOutputManager> getRegistry()
   {
      return registry;
   }

   /**
    * Set status text message.
    * 
    * @param   message
    *          Text message.
    */
   public void setStatusText(String message)
   {
      GuiDriver screen = (GuiDriver) driver;
      screen.setStatusText(message);
   }   
   
   /**
    * Set status text message of specified window.
    * 
    * @param   message
    *          Text message.
    * @param   window
    *          Window instance.
    */
   public void setStatusText(Window<?> window, String message)
   {
      GuiDriver screen = (GuiDriver) driver;
      screen.setStatusText(window, message);      
   }   

   /**
    * Check if the given key can raise the STOP condition.
    * 
    * @param    key
    *           The key code to be checked.
    * 
    * @return   <code>true</code> if the key is the {@link Key#VK_CTRL_BREAK} key.
    */
   @Override
   protected boolean raiseStopCondition(int key)
   {
      return key == Key.VK_CTRL_BREAK;
   }
   
   /**
    * Check if the window being drawn needs to be overridden.
    * <p>
    * In GUI, the <code>window</code> parameter is returned.
    *  
    * @param    window
    *           The currently active window.
    *           
    * @return   See above.
    */
   protected TitledWindow<?> overrideWindow(Window<?> window)
   {
      return window;
   }

   /**
    * Enable or Disable OS-level events for the specified windows.
    * 
    * @param    windowIds
    *           The window IDs which need to be processed.
    * @param    capture
    *           Flag indicating OS level events are captured (when <code>true</code>).
    */
   public void enableOSEvents(int[] windowIds, boolean capture)
   {
      ((GuiDriver) driver).enableEvents(windowIds, capture);
   }
}
