/*
** Module   : GuiPrimitivesImpl.java
** Abstract : generic implementation of GUI primitives
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140523 First version.
** 002 CA  20140711 Enhanced font support.
** 003 CA  20140806 Added misc APIs related to window behavior/state.
** 004 MAG 20140905 Added fillPoly method.
** 005 HC  20140911 GUI coordinate support - data types refactoring.
** 006 MAG 20140925 Check if window has renderer.
** 007 CA  20141021 Changed the drawImage API to receive the image's width and height.
** 008 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 009 EVL 20141116 Adding round rectangles draw and fill.
** 010 EVL 20141125 Adding dotted line draw capability.
** 011 CA  20141220 Added APIs to allow text drawing and font selection matching legacy sizes.
** 012 HC  20141207 Implemented runtime support for several SESSION attributes.
** 013 EVL 20150203 Adding image offset drawing capability.
** 014 EVL 20150218 Adding support TRANSPARENT, STRETCH-TO-FIT, RETAIN-SHAPE and CONVERT-3D-COLORS
**                  options to image drawing.
** 015 CA  20150317 Added API to determine the font's max width.
** 016 CA  20150408 Added APIs to set the OS window's title and to de-iconify the window.
** 017 CA  20150507 Added APIs to draw text centered vertically.
** 018 GES 20150501 Moved virtually all functionality out of this class and into the
**                  AbstractGuiDriver, the SwingGuiDriver or to the new abstract parent class.
**                  The remainder is now generic code that can be shared between all GUI clients.
*/

package com.goldencode.p2j.ui.client.gui.driver;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;

/**
 * This is a low level GUI-specific window emulator management implementation.  Real drawing does
 * not occur in this class. This is a kind of router that keeps the list of emulated window
 * instances and a current instance which can be swapped.
 * <p>
 * This interface is for GUI-specific features. It is intended to be specific to a given
 * UI-modality (GUI in this case). No driver-specific implementation classes or features
 * should be present.  For example, no Swing-specific feature usage is allowed.
 * <p>
 * The actual GUI processing does not heavily utilize the pluggable <code>OutputPrimitives</code>
 * interface.  In other words, the interactive UI processing GUI is not switched in and out like
 * in ChUI.  For this reason, the core output processing in ChUI is driven using the primitives
 * approach, but in GUI we simply honor the minimum implementation in this class.  The majority
 * of this class is really about providing access to and management of the core window structure.
 * The primitives implementation is secondary, but the functionality is best implemented here.
 */
public abstract class GuiPrimitivesImpl
implements GuiPrimitives
{
   /** Current GUI emulator. */
   protected EmulatedWindowState ews = null;
   
   /** Registry of window emulators. */
   private Map<Integer, EmulatedWindowState> emulators = new HashMap<>();
   
   /**
    * Constructor.
    */
   public GuiPrimitivesImpl()
   {
   }

   /**
    * Clear specified area at the screen.
    * 
    * @param    left
    *           Left coordinate of the area.
    * @param    top
    *           Top coordinate of the area.
    * @param    right
    *           Right coordinate of the area.
    * @param    bottom
    *           Bottom coordinate of the area.
    * @param    color
    *           Color.
    */
   @Override
   public void clearArea(int left, int top, int right, int bottom, Color color)
   {
      // no-op
   }

   /**
    * Refresh specified area.
    * <p>
    * This is a no-op.
    * 
    * @param    rect
    *           Area to refresh. 
    */
   @Override
   public void refresh(NativeRectangle rect)
   {                                                                            
      // no-op
   }

   /**
    * Get the screen width, in character units.
    * 
    * @return    See above.
    */
   @Override
   public int screenWidth()
   {
      return ews.screenWidth();
   }

   /**
    * Get the screen height, in character units.
    * 
    * @return    See above.
    */
   @Override
   public int screenHeight()
   {
      return ews.screenHeight();
   }
   
   /**
    * Synchronize screen.
    * <p>
    * Driver level implementation of this method should force flushing internal
    * low-level buffers to screen. This method requires implementation only if
    * operating system or hardware uses internal buffering for the terminal 
    * output. Otherwise this method can do nothing. 
    */
   @Override
   public abstract void sync();
   
   /**
    * Set the active screen bitmap for the current window.
    * 
    * @param    bitmap
    *           The screen bitmap.
    */
   @Override
   public void setScreenBitmap(ScreenBitmap bitmap)
   {
      ews.setScreenBitmap(bitmap);
   }

   /**
    * Get the screen bitmap for the current window.
    * 
    * @return   The current window's screen bitmap.
    */
   @Override
   public ScreenBitmap getScreenBitmap()
   {
      return ews.getScreenBitmap();
   }

   /**
    * Get the window ID for the current emulated window.
    */
   @Override
   public int getWindowId()
   {
      return ews.getWindowId();
   }
   
   /**
    * Get the currently selected window emulator.
    * 
    * @return   See above.
    */
   @Override
   public EmulatedWindowState getWindowEmulator()
   {
      return ews;
   }
   
   /**
    * Returns the window emulator registered for the supplied window id.
    * 
    * @param    windowId
    *           A valid window id.
    *           
    * @return   The identified window emulator.
    * 
    * @throws   RuntimeException
    *           When no window emulator is registered for the supplied window id.
    */
   @Override
   public EmulatedWindowState getWindowEmulator(int windowId)
   {
      EmulatedWindowState ews = emulators.get(windowId);
      
      if (ews == null)
      {
         throw new RuntimeException("No renderer is registered for id = " + windowId);
      }

      return ews;
   }
   
   /**
    * Check if window has an emulator for the given window id.
    * 
    * @param    windowId
    *           Window ID.
    *           
    * @return   <code>true</code> if window has an associated emulator, <code>true</code>
    *           otherwise.
    */
   @Override
   public boolean hasWindowEmulator(int windowId)
   {
      return emulators.get(windowId) != null;
   }

   /**
    * Register a new window with the driver.
    * 
    * @param    windowId
    *           The window ID.
    * @param    ews
    *           New emulated window state to register.  
    */
   @Override
   public void registerWindow(int windowId, EmulatedWindowState ews)
   {
      if (ews == null)
      {
         throw new NullPointerException("Null simulator is not allowed!");
      }
      
      emulators.put(windowId, ews);
      
      ews.setWindowId(windowId);
   }

   /**
    * Deregister this window from the driver.  This needs to be called only when the 
    * window widget is being destroyed.
    * 
    * @param    windowId
    *           The window ID.
    */
   @Override
   public void deregisterWindow(int windowId)
   {
      EmulatedWindowState ews = emulators.remove(windowId);
      ews.quit();
   }
   
   /**
    * On a GUI interface multiple windows can be created. Each window has it own emulator.
    * When a window is first drawn, this method is used to set it's emulator as the current 
    * target for output operations.  
    * 
    * @param    windowId
    *           Window ID.
    */
   @Override
   public void selectWindow(int windowId)
   {
      ews = getWindowEmulator(windowId);
   }
   
   /**
    * Perform shutdown.
    */
   @Override
   public void shutdown()
   {
      for (EmulatedWindowState ews : emulators.values())
      {
         ews.quit();
      }
      
      emulators.clear();
   }
}
