/*
** Module   : MouseResizeable.java
** Abstract : Implementation of a mouse resizeable widget.
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140805 Created initial version.
** 002 HC  20140911 GUI coordinate support - data types refactoring.
** 003 CA  20141021 Removed the window-specific code.  This class is now generic enough to be used
**                  by any widget.
** 004 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 005 HC  20150323 WidgetId.getId() renamed to asInt().
** 006 GES 20150501 Removed AWT dependencies and moved to a more generic package.
*/

package com.goldencode.p2j.ui.client.gui.driver;
                                   
// this usage of AWT events (and Point and Rectangle) is intentional, they are being used
// separately from the AWT/Swing event model; no other AWT/Swing usage is occurring in this class
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.*;

import com.goldencode.p2j.ui.client.gui.*;
                                                      
/**
 * Listener for mouse actions which can trigger a widget to be resized.
 */
public class MouseResizeable
extends MouseWidgetAction                               
{
   /** Flag indicating if resize is currently active. */
   private boolean canResize = false;
   
   /** The last resize-related mouse event. */
   private MouseEvent resizeEvent;

   /** The active mouse cursor type. */
   private CursorType activeCursor;
   
   /**
    * Create a new mouse resizeable action, for the specified widget.
    * 
    * @param    wid
    *           The widget ID.
    */
   public MouseResizeable(int wid)
   {
      super(wid);
   }

   /**
    * Notification of a mouse pressed event for this {@link MouseWidgetAction#widget}.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      if (canResize && e.getButton() == MouseEvent.BUTTON1)
      {
         // dragging has started - save the position
         resizeEvent = e;
      }
   }

   /**
    * Notification of a mouse released event for this {@link MouseWidgetAction#widget}.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      if (canResize)
      {
         resizeEvent = null;
      }
   }

   /**
    * Notification of a mouse dragged event for this {@link MouseWidgetAction#widget}.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseDragged(MouseEvent e)
   {                                                  
      int id = (resizeEvent == null ? -1 : resizeEvent.getID());
      
      if (canResize && (id == MouseEvent.MOUSE_DRAGGED || id == MouseEvent.MOUSE_PRESSED))
      {
         // check which cursor is in use
         boolean setLocationX = activeCursor == CursorType.NW_RESIZE ||
                                activeCursor == CursorType.W_RESIZE  ||
                                activeCursor == CursorType.SW_RESIZE;
         boolean setLocationY = activeCursor == CursorType.NW_RESIZE  ||
                                activeCursor == CursorType.N_RESIZE   ||
                                activeCursor == CursorType.NE_RESIZE;
         
         // depending on the dragging point, it will resize x axis, y axis or both
         boolean resizeX = activeCursor == CursorType.NW_RESIZE || 
                           activeCursor == CursorType.NE_RESIZE ||
                           activeCursor == CursorType.SW_RESIZE ||
                           activeCursor == CursorType.SE_RESIZE ||
                           activeCursor == CursorType.E_RESIZE  ||
                           activeCursor == CursorType.W_RESIZE;
         boolean resizeY = activeCursor == CursorType.NW_RESIZE || 
                           activeCursor == CursorType.NE_RESIZE ||
                           activeCursor == CursorType.SW_RESIZE ||
                           activeCursor == CursorType.SE_RESIZE ||
                           activeCursor == CursorType.N_RESIZE  ||
                           activeCursor == CursorType.S_RESIZE;
         int signX = activeCursor == CursorType.W_RESIZE ||
                     activeCursor == CursorType.NW_RESIZE ||
                     activeCursor == CursorType.SW_RESIZE ? 1 : -1;
         int signY = activeCursor == CursorType.N_RESIZE ||
                     activeCursor == CursorType.NW_RESIZE ||
                     activeCursor == CursorType.NE_RESIZE ? 1 : -1;
   
         int dx = resizeX ? e.getXOnScreen() - resizeEvent.getXOnScreen() : 0;
         int dy = resizeY ? e.getYOnScreen() - resizeEvent.getYOnScreen() : 0;
         
         dx = dx * signX;
         dy = dy * signY;
         
         WindowGuiImpl window = (WindowGuiImpl) widget.ancestor();
         gd.selectWindow(window.config().id.asInt());

         gd.startBatch();
         
         try
         {
            widget.resize(dx, dy, setLocationX, setLocationY);
   
            resizeEvent = e;
         }
         finally
         {
            // window might have changed if a trigger was invoked...
            gd.selectWindow(window.config().id.asInt());

            gd.endBatch();
         }
      }
   }

   /**
    * Notification of a mouse exited event for this {@link MouseWidgetAction#widget}.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseExited(MouseEvent e)
   {
      if (resizeEvent == null)
      {
         WindowGuiImpl window = (WindowGuiImpl) widget.ancestor();
         gd.selectWindow(window.config().id.asInt());

         gd.setCursor(CursorType.DEFAULT);
         activeCursor = CursorType.DEFAULT;           
         canResize = false;
      }
   }

   /**
    * Notification of a mouse moved event for this {@link MouseWidgetAction#widget}.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseMoved(MouseEvent e)
   {
      Point p = e.getLocationOnScreen();
      com.goldencode.p2j.ui.client.NativeRectangle bounds = widget.physicalBounds();
      
      // we have 8 cases; corners have precedence
      CursorType[] cursors =
      {
         CursorType.NW_RESIZE,
         CursorType.NE_RESIZE,
         CursorType.SE_RESIZE,
         CursorType.SW_RESIZE,
         CursorType.N_RESIZE,
         CursorType.E_RESIZE,
         CursorType.S_RESIZE,
         CursorType.W_RESIZE,
      };
      int[] xcoords1 =                         
      {
         bounds.left(),  // NW
         bounds.right(), // NE
         bounds.right(), // SE
         bounds.left(),  // SW
         bounds.left(),  // N
         bounds.right(), // E
         bounds.left(),  // S
         bounds.left(),  // W
      };
      int[] xcoords2 =
      {
         bounds.left(),  // NW
         bounds.right(), // NE
         bounds.right(), // SE
         bounds.left(),  // SW
         bounds.right(), // N
         bounds.right(), // E
         bounds.right(), // S
         bounds.left(),  // W
      };
      int[] ycoords1 =
      {
         bounds.top(),    // NW
         bounds.top(),    // NE
         bounds.bottom(), // SE
         bounds.bottom(), // SW
         bounds.top(),    // N
         bounds.top(),    // E
         bounds.bottom(), // S
         bounds.top(),    // W
      };
      int[] ycoords2 =
      {
         bounds.top(),    // NW
         bounds.top(),    // NE
         bounds.bottom(), // SE
         bounds.bottom(), // SW
         bounds.top(),    // N
         bounds.bottom(), // E
         bounds.bottom(), // S
         bounds.bottom(), // W
      };
      
      CursorType cursor = CursorType.DEFAULT;
      canResize = false;
      for (int i = 0; i < cursors.length; i++)
      {
         int x1 = xcoords1[i] - GuiConstants.BORDER_SIZE;
         int y1 = ycoords1[i] - GuiConstants.BORDER_SIZE;
         int x2 = xcoords2[i] + GuiConstants.BORDER_SIZE;
         int y2 = ycoords2[i] + GuiConstants.BORDER_SIZE;
         
         Rectangle r = new Rectangle(x1, y1, x2 - x1 + 1, y2 - y1 + 1);
         if (r.contains(p))
         {
            cursor = cursors[i];
            canResize = true;
            break;
         }
      }
      
      WindowGuiImpl window = (WindowGuiImpl) widget.ancestor();
      gd.selectWindow(window.config().id.asInt());

      if (cursor != activeCursor)
      {
         gd.setCursor(cursor);
         activeCursor = cursor;
      }
      else if (!canResize)
      {
         cursor = CursorType.DEFAULT;
         gd.setCursor(cursor);
         activeCursor = cursor;
      }
   }
}
