/*
** Module   : SwingEmulatedWindow.java
** Abstract : core GUI renderer for AWT/Swing client
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140508 First version.
** 002 CA  20140711 Enhanced font support - removed fontSize.
** 003 CA  20140806 Added mouse support and other misc APIs related to window state.
** 004 MAG 20140912 Added fill polygon.
** 005 HC  20140911 GUI coordinate support - data types refactoring.
** 006 MAG 20140924 Create graphical object when buffered image is created.
** 007 CA  20141021 Added notification of OS window events.  Added support for batch drawing: the 
**                  current AWT repaint processing will wait until the last paint primitive is
**                  received, before returning the control to the AWT event processing thread, to
**                  avoid flickering.  Do not allow window maximize via OS means.
**                  Added implementation for new paint primitives: SET_LOCATION, RESIZE_WINDOW.
** 008 HC  20141031 Implemented support for drawing highlighted areas.
** 009 MAG 20141020 Improve clipping rectangle management. Added mouse wheel listener.
** 010 CA  20141029 Each simulator needs to receive the swing keyboard reader implementation in 
**                  use.  It's the callers responsibility to ensure that the same instance is used
**                  for all simulators, if the client is ran in a multi-window mode (i.e. GUI).
** 011 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 012 EVL 20141116 Adding round rectangles draw and fill.
** 013 CA  20141123 Fixed processing of window focus gained/lost.  Added APIs to allow movement of
**                  a window to top (at OS level).  When the window is destroyed, first remove the
**                  listeners, so no key/mouse/window events is processed while the window is being
**                  destroyed.
** 014 CA  20141125 Do not accept negative dimensions, when resizing the screen buffers.
**                  Added API to conditionally enable/disable OS-level events.
** 015 EVL 20141125 Adding dotted line draw capability.
** 016 CA  20141220 Added APIs to allow text drawing and font selection matching legacy sizes.
** 017 HC  20141207 Implemented runtime support for several SESSION attributes.
** 018 HC  20150106 Javadoc fixes.
** 019 EVL 20150203 Adding image offset drawing capability.
** 020 EVL 20150218 Adding transparency handling.
** 021 EVL 20150509 Fixing clipping rectangle dimension computing.
** 022 CA  20150408 All window events are treated as OS events.  Allow the title of the OS window
**                  to be set.  Added deiconifyWindow and isIconified.
** 023 CA  20150429 Fixed a problem in setScreenBitmap, when the received bitmap is not yet aware
**                  of the window's dimensions.
** 024 CA  20150507 Allow text to becentered vertically.
**                  Moved from legacy metric scaling to DPI scaling from 72 to OS DPI.
** 025 GES 20150325 Major refactoring to move all generic functionality to a new abstract parent
**                  class or to the new abstract GUI driver. Old JComponent parent is now just
**                  contained as a member.
*/

package com.goldencode.p2j.ui.client.gui.driver.swing;

import java.awt.*;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.event.*;
import java.awt.event.WindowEvent;
import java.awt.font.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.lang.reflect.*;
import java.util.*;
import java.util.List;
import java.util.concurrent.*;

import javax.swing.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.driver.swing.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.driver.*;

/**
 * The core GUI renderer. Because the painting results should be permanent a BufferedImage is
 * used to draw and keep the last screen image. A pool of drawing operations is used to queue 
 * drawing primitives. The operations are enqueued in the pool one at a time and then are drawn
 * on the buffered image. When the pool is empty, the screen is updated with the image from the
 * stored buffered image.  
 */
public class SwingEmulatedWindow 
extends EmulatedWindowState<Font, BufferedImage>
{
   /** Map of saved event listeners, which can be restored at a later time. */
   private Map<Class<? extends EventListener>, List<EventListener>> savedListeners = null;
   
   /** Content pane for the window. */
   private ContentPane pane = null;

   /** Screen buffer */
   private BufferedImage screenBuffer;
   
   /** Graphical object */
   private Graphics2D g2;

   /** A stack of clipping rectangles. */
   private Deque<java.awt.Rectangle> clippings = new ConcurrentLinkedDeque<>();

   /** The X translated location of the drawing area, on the physical canvas. */
   private int translateX = 0;
   
   /** The Y translated location of the drawing area, on the physical canvas. */
   private int translateY = 0;
   
   /** Screen buffer for clearing an active highlight */
   private BufferedImage highlightRestoreBuffer;
   
   /** The left location of the last highlighted widget. */
   private int highlightX;
   
   /** The top location of the last highlighted widget. */
   private int highlightY;

   /** The window focus listeners for this simulator. */
   private WindowFocusListener[] savedFocusListeners;
   
   /** Containing frame (may be an applet or an application). */
   private Frame frame = null;
   
   /** Dotted line pattern. */
   private Stroke dotStroke;
   
   /** Small dotted line pattern. */
   private Stroke dotSmallStroke;
   
   /** Small and thin dotted line pattern. */
   private Stroke dotSmallThinStroke;
   
   /** Temporary context stored during <code>paintComponent()</code> operations. */
   private Graphics graphics = null;
   
   /**
    * Constructor.
    * 
    * @param   frame
    *          Application container.
    * @param   keyReader
    *          The {@link SwingKeyboardReader key reader} which listens for incoming keys.
    */
   public SwingEmulatedWindow(JFrame frame, SwingKeyboardReader keyReader)
   {
      pane = new ContentPane(frame, keyReader)
      {
         public void paintComponent(Graphics graphics)
         {
            super.paintComponent(graphics);
            
            // save off graphics context (this is needed because other GUI clients don't use
            // the same graphics code and drawOnScreenBitmap() is common code so it can't be
            // passed as a parameter)
            SwingEmulatedWindow.this.graphics = graphics;
            
            SwingEmulatedWindow.this.drawOnScreenBitmap();
            
            // clear graphics context
            SwingEmulatedWindow.this.graphics = null;
         }   
      };

      frame.setUndecorated(true);
      
      this.frame = frame;

      // register listeners for OS window events
      final ThinClient tc = ThinClient.getInstance();
      frame.addWindowListener(new WindowAdapter()
      {
         @Override
         public void windowIconified(WindowEvent e)
         {
            iconifyWindow(-1);
         }
         
         @Override
         public void windowDeiconified(WindowEvent e)
         {
            deiconifyWindow(-1);
         }
      });
      
      frame.addWindowFocusListener(new WindowFocusListener()
      {
         @Override                                  
         public void windowLostFocus(WindowEvent e)
         {
            // a LEAVE event is not raised; it will be raised by the 4GL window which received
            // the focus.  If focus is received by another OS window, then no LEAVE needs to be
            // sent.
            
            WindowGuiImpl window = (WindowGuiImpl) WindowManager.findWindow(windowId);

            boolean toOS = (e.getOppositeWindow() == null);
            
            if (toOS)
            {
               // if focus is lost to a non-4GL window, the titlebar needs to be aware of this
               ThinClient.getInstance().postOSEvent(new WindowActivated(window, false, toOS));
            }
         }
         
         @Override
         public void windowGainedFocus(WindowEvent e)
         {
            WindowGuiImpl window = (WindowGuiImpl) WindowManager.findWindow(windowId);

            boolean fromOS = (e.getOppositeWindow() == null);
            ThinClient.getInstance().postOSEvent(new WindowActivated(window, true, fromOS));
         }
      });

      frame.addWindowStateListener(new WindowStateListener()
      {
         public void windowStateChanged(WindowEvent e)
         {
            if ((e.getNewState() & Frame.MAXIMIZED_BOTH) == Frame.MAXIMIZED_BOTH)
            {
               resizeWindow(pane.getWidth(), pane.getHeight());
            }
         }
      });
   
      // setup mouse listeners
      pane.addMouseListener(mouseHandler);
      pane.addMouseMotionListener(mouseHandler);
      pane.addMouseWheelListener(mouseHandler);
      
      // initialize dot pattern 
      dotStroke = new BasicStroke(LineStroke.DOT_LINE_WIDTH,
                                  BasicStroke.CAP_BUTT,
                                  BasicStroke.JOIN_BEVEL,
                                  0.0f,
                                  new float[] {LineStroke.DOT_STEP, LineStroke.DOT_STEP},
                                  LineStroke.DOT_STEP);
      
      dotSmallStroke = new BasicStroke(LineStroke.DOT_LINE_WIDTH,
                                       BasicStroke.CAP_BUTT,
                                       BasicStroke.JOIN_BEVEL,
                                       0.0f,
                                       new float[] {LineStroke.DOT_STEP_SMALL,
                                                    LineStroke.DOT_STEP_SMALL},
                                       LineStroke.DOT_STEP_SMALL);
      
      dotSmallThinStroke = new BasicStroke(LineStroke.DOT_LINE_THIN_WIDTH,
                                           BasicStroke.CAP_BUTT,
                                           BasicStroke.JOIN_BEVEL,
                                           0.0f,
                                           new float[] {LineStroke.DOT_STEP_SMALL,
                                                        LineStroke.DOT_STEP_SMALL},
                                           LineStroke.DOT_STEP_SMALL);
      // TODO: disable the popup of the window actions (min/max/close/etc) via key? 
   }

   /**
    * Cause the outer frame to be disposed, as the window is being terminated.
    */
   @Override
   public void quit()
   {
      captureMouseEvents(false);

      for (MouseListener l : pane.getMouseListeners())
      {
         pane.removeMouseListener(l);
      }
      for (MouseMotionListener l : pane.getMouseMotionListeners())
      {
         pane.removeMouseMotionListener(l);
      }
      for (MouseWheelListener l : pane.getMouseWheelListeners())
      {
         pane.removeMouseWheelListener(l);
      }
      for (WindowListener l : frame.getWindowListeners())
      {
         frame.removeWindowListener(l);
      }
      for (WindowFocusListener l : frame.getWindowFocusListeners())
      {
         frame.removeWindowFocusListener(l);
      }
      for (WindowStateListener l : frame.getWindowStateListeners())
      {
         frame.removeWindowStateListener(l);
      }
      for (KeyListener l : frame.getKeyListeners())
      {
         frame.removeKeyListener(l);
      }
   }
   
   /**
    * Cause repaint processing to occur.
    */
   @Override
   public void repaint()
   {
      pane.repaint();
   }

   /**
    * Hide or show the window, depending on the {@code visible} flag.
    * 
    * @param    visible
    *           The window visibility state.
    */
   @Override
   public void setWindowVisible(boolean visible)
   {
      frame.pack();
      frame.setVisible(visible);
   }
   
   /**
    * Check if the simulated window is visible.
    * 
    * @return   See above.
    */
   @Override
   public boolean isWindowVisible()
   {
      return frame.isVisible();
   }

   /** 
    * Get the native screen size, in pixel units.
    * 
    * @return   See above.
    */
   @Override
   public NativeDimension getScreenSize()
   {
      Dimension dim = Toolkit.getDefaultToolkit().getScreenSize();
      NativeDimension ndim = new NativeDimension(dim.width, dim.height);
      
      return ndim;
   }

   /**
    * Get the window location on screen.
    * 
    * @return   This window's location.
    */
   @Override
   public NativePoint getWindowLocation()
   {
      java.awt.Point p = frame.getLocation();
      
      return new NativePoint(p.x, p.y);
   }

   /**
    * Get the window height.
    * 
    * @return   This window's width in pixels.
    */
   @Override
   public int getHeight()
   {
      return pane.getHeight();
   }
   
   /**
    * Get the window width.
    * 
    * @return   This window's width in pixels.
    */
   @Override
   public int getWidth()
   {
      return pane.getWidth();
   }
   
   /**
    * Manage any window state related to iconification.
    * 
    * @param    iconify
    *           <code>true</code> if this window is being iconified, <code>false</code> for
    *           de-iconified.
    */
   @Override
   public void setIconificationState(boolean iconify)
   {
      super.setIconificationState(iconify);
      
      if (iconify)
      {
         frame.setState(Frame.ICONIFIED);
      }
      else
      {
         int extendedState = frame.getExtendedState() & ~Frame.ICONIFIED;
         frame.setExtendedState(extendedState);
      }
   }

   /**
    * Set the mouse cursor to the given type.
    * 
    * @param    type
    *           The cursor type.
    */
   @Override
   public void setCursor(CursorType type)
   {
      int curs = Cursor.DEFAULT_CURSOR;
      
      switch (type)
      {
         case TEXT:
            curs = Cursor.TEXT_CURSOR;
            break;
         case N_RESIZE:
            curs = Cursor.N_RESIZE_CURSOR;
            break;
         case NE_RESIZE:
            curs = Cursor.NE_RESIZE_CURSOR;
            break;
         case E_RESIZE:
            curs = Cursor.E_RESIZE_CURSOR;
            break;
         case SE_RESIZE:
            curs = Cursor.SE_RESIZE_CURSOR;
            break;
         case S_RESIZE:
            curs = Cursor.S_RESIZE_CURSOR;
            break;
         case SW_RESIZE:
            curs = Cursor.SW_RESIZE_CURSOR;
            break;
         case W_RESIZE:
            curs = Cursor.W_RESIZE_CURSOR;
            break;
         case NW_RESIZE:
            curs = Cursor.NW_RESIZE_CURSOR;
            break;
      }
      
      pane.setCursor(Cursor.getPredefinedCursor(curs));
   }

   /**
    * Save (to {@link #savedFocusListeners} and remove the OS window's focus listeners.
    */
   @Override
   public void saveFocusListeners()
   {
      WindowFocusListener[] flisteners = frame.getWindowFocusListeners();
      for (WindowFocusListener l : flisteners)
      {
         frame.removeWindowFocusListener(l);
      }
      
      this.savedFocusListeners = flisteners;
   }

   /**                                                 
    * Restore the OS window's focus listeners, from the {@link #saveFocusListeners()} list.
    */
   @Override
   public void restoreFocusListeners()
   {
      for (WindowFocusListener l : savedFocusListeners)
      {
         frame.addWindowFocusListener(l);
      }
      this.savedFocusListeners = null;
   }

   /**
    * Move this window to top, at the OS level.
    */
   @Override
   public void moveToTop()
   {
      if (!isWindowVisible())
      {
         return;
      }
   
      try
      {
         SwingUtilities.invokeAndWait(new Runnable()
         {
            @Override
            public void run()
            {
               frame.setAlwaysOnTop(true);
               frame.toFront();
               frame.requestFocus();
               frame.setAlwaysOnTop(false);
            }
         });
      }
      catch (InvocationTargetException | InterruptedException e)
      {
         throw new RuntimeException(e);
      }
   }
   
   /**
    * Determine if the window is on top of all windows, including OS.
    * 
    * @return   See above.
    */
   @Override
   public boolean isOnTop()
   {
      return frame.isActive();
   }

   /**
    * Enable or Disable OS-level events for the specified windows.
    * 
    * @param    capture
    *           Flag indicating OS level events are captured (when <code>true</code>).
    */
   @Override
   public void enableEvents(boolean capture)
   {
      if (capture)
      {
         if (savedListeners == null)
         {
            return;
         }
         
         List<EventListener> events;
         
         events = savedListeners.remove(MouseListener.class);
         for (EventListener l : events)
         {
            pane.addMouseListener((MouseListener) l);
         }
         
         events = savedListeners.remove(MouseMotionListener.class);
         for (EventListener l : events)
         {
            pane.addMouseMotionListener((MouseMotionListener) l);
         }
   
         events = savedListeners.remove(MouseWheelListener.class);
         for (EventListener l : events)
         {
            pane.addMouseWheelListener((MouseWheelListener) l);
         }
   
         events = savedListeners.remove(WindowListener.class);
         for (EventListener l : events)
         {
            frame.addWindowListener((WindowListener) l);
         }
   
         events = savedListeners.remove(WindowFocusListener.class);
         for (EventListener l : events)
         {
            frame.addWindowFocusListener((WindowFocusListener) l);
         }
   
         events = savedListeners.remove(WindowStateListener.class);
         for (EventListener l : events)
         {
            frame.addWindowStateListener((WindowStateListener) l);
         }
   
         events = savedListeners.remove(KeyListener.class);
         for (EventListener l : events)
         {
            frame.addKeyListener((KeyListener) l);
         }
         
         savedListeners = null;
      }
      else
      {
         if (savedListeners != null)
         {
            return;
         }
   
         savedListeners = new HashMap<>();
   
         List<EventListener> events;
         
         events = new ArrayList<>();
         for (MouseListener l : pane.getMouseListeners())
         {
            pane.removeMouseListener(l);
            events.add(l);
         }
         savedListeners.put(MouseListener.class, Collections.unmodifiableList(events));
         
         events = new ArrayList<>();
         for (MouseMotionListener l : pane.getMouseMotionListeners())
         {
            pane.removeMouseMotionListener(l);
            events.add(l);
         }
         savedListeners.put(MouseMotionListener.class, Collections.unmodifiableList(events));
   
         events = new ArrayList<>();
         for (MouseWheelListener l : pane.getMouseWheelListeners())
         {
            pane.removeMouseWheelListener(l);
            events.add(l);
         }
         savedListeners.put(MouseWheelListener.class, Collections.unmodifiableList(events));
   
         events = new ArrayList<>();
         for (WindowListener l : frame.getWindowListeners())
         {
            frame.removeWindowListener(l);
            events.add(l);
         }
         savedListeners.put(WindowListener.class, Collections.unmodifiableList(events));
   
         events = new ArrayList<>();
         for (WindowFocusListener l : frame.getWindowFocusListeners())
         {
            frame.removeWindowFocusListener(l);
            events.add(l);
         }
         savedListeners.put(WindowFocusListener.class, Collections.unmodifiableList(events));
   
         events = new ArrayList<>();
         for (WindowStateListener l : frame.getWindowStateListeners())
         {
            frame.removeWindowStateListener(l);
            events.add(l);
         }
         savedListeners.put(WindowStateListener.class, Collections.unmodifiableList(events));
   
         events = new ArrayList<>();
         for (KeyListener l : frame.getKeyListeners())
         {
            frame.removeKeyListener(l);
            events.add(l);
         }
         savedListeners.put(KeyListener.class, Collections.unmodifiableList(events));
      }
   }

   /**
    * Get the metrics of the given font.
    *  
    * @param    f
    *           The font for which the font metrics are to be obtained.
    *           
    * @return   The font metrics.
    */
   @Override
   public FontMetricsHelper getFontMetrics(FontDetails<Font> f)
   {
      synchronized (lock())
      {
         return new SwingFontMetrics(getNativeFontMetrics(f.font));
      }
   }

   /**
    * Get the height of the specified text, using the given font.
    * 
    * @param    text
    *           The text.
    * @param    f
    *           The font used to draw and measure the text.
    *           
    * @return   The text's height.
    */
   @Override
   public int getTextHeight(String text, FontDetails<Font> f)
   {
      synchronized (lock())
      {
         FontMetrics fm = getNativeFontMetrics(f.font);
         LineMetrics lm = fm.getLineMetrics(text, g2 == null ? pane.getGraphics() : g2);
         
         return (int) lm.getHeight();
      }
   }
   
   /**
    * Draw using a paint operation.
    * 
    * @param    ps
    *           Paint operation parameters.
    */
   @Override
   public void draw(PaintStructure<Font, BufferedImage> ps)
   {
      switch (ps.id)
      {
         case DRAW_STRING:
            drawString(ps);
            break;
         case DRAW_STRING_SCALED:
            drawStringScaled(ps);
            break;
         case DRAW_LINE:
            g2.drawLine(ps.x, ps.y, ps.width, ps.height);
            break;
         case DRAW_RECT:
            g2.drawRect(ps.x, ps.y, ps.width, ps.height);
            break;
         case DRAW_ROUND_RECT:
            g2.drawRoundRect(ps.x, ps.y, ps.width, ps.height, ps.arcDiameter, ps.arcDiameter);
            break;
         case DRAW_IMAGE:
            BufferedImage img2draw = null;
            int width  = 0;
            int height = 0;
            // scalng factors for vertical and horizontal
            float kHor = 0;
            float kVer = 0;
            if (ps.xOffset == -1 && ps.yOffset == -1)
            {
               // full original image
               img2draw = ((ImageWrapper<BufferedImage>)ps.img).getImage();
               width    = ps.width;
               height   = ps.height;
               kHor = ((float)ps.width) / ((float) ps.img.getWidth());
               kVer = ((float)ps.height) / ((float) ps.img.getHeight()); 
            }
            else
            {
               // need to recalc the width and height for subimage to cut
               // to not go out of image borders
               int widthToCut = ps.img.getWidth() - ps.xOffset;
               if (widthToCut > ps.width)
               {
                  widthToCut = ps.width;
               }
               int heightToCut = ps.img.getHeight() - ps.yOffset;
               if (heightToCut > ps.height)
               {
                  heightToCut = ps.height;
               }
               // partial image, get subimage
               BufferedImage fullImg = ((ImageWrapper<BufferedImage>)ps.img).getImage();
               BufferedImage bimg = fullImg.getSubimage(ps.xOffset,
                                                        ps.yOffset,
                                                        widthToCut,
                                                        heightToCut);
               
               if (bimg != null)
               {
                  img2draw = bimg;
                  width    = widthToCut;
                  height   = heightToCut;
                  kHor = ((float)ps.width) / ((float)widthToCut);
                  kVer = ((float)ps.height) / ((float)heightToCut); 
               }
            }
            // finally draw the image
            if (img2draw != null)
            {
               // check binary image options
               // 3D color conversions
               if (ps.convert3D)
               {
                  img2draw = get3DColorsConvertedImage((BufferedImage)img2draw);
               }
               // transparency
               if (ps.transparent)
               {
                  int pixTrans = img2draw.getRGB(0, height - 1);
                  img2draw = getTransparentImage(img2draw, pixTrans);
               }
               // resizing
               if (ps.stretchToFit)
               {
                  if (ps.retainShape)
                  {
                     // need to calculate native aspect ratio for image
                     // if we have two different scaling factors for x and y we choose only one
                     // to keep scaling in both directions, this must be one that is smaller
                     if (kHor < kVer)
                     {
                        width  = ps.width;
                        height = Math.round(((float)height) * kHor);
                     }
                     else
                     {
                        width = Math.round(((float)width) * kVer);
                        height = ps.height;
                     }
                  }
                  else
                  {
                     width  = ps.width;
                     height = ps.height;
                  }
               }
               // call Java backend
               g2.drawImage(img2draw, ps.x, ps.y, width, height, pane);
            }
            break;
         case FILL_RECT:
            g2.fillRect(ps.x, ps.y, ps.width, ps.height);
            break;
         case FILL_ROUND_RECT:
            g2.fillRoundRect(ps.x, ps.y, ps.width, ps.height, ps.arcDiameter, ps.arcDiameter);
            break;
         case FILL_POLYGON:
            g2.fillPolygon(ps.xPoints, ps.yPoints, ps.nPoints);
            break;
         case DRAW_3D_RECT:
            g2.draw3DRect(ps.x, ps.y, ps.width, ps.height, ps.raised);
            break;
         case FILL_3D_RECT:
            g2.fill3DRect(ps.x, ps.y, ps.width, ps.height, ps.raised);
            break;
         case SET_COLOR:
            g2.setColor(SwingHelper.fromColorRgb(ps.color));
            break;
         case SET_LINE_STROKE:
            // default line stroke
            Stroke stroke = new BasicStroke();
            
            switch (ps.stroke)
            {
               case SPECIFIED_WIDTH:
                  stroke = new BasicStroke(ps.strokeWidth);
                  break;
               case DOTTED:
                  stroke = dotStroke;
                  break;
               case DOTTED_SMALL:
                  stroke = dotSmallStroke;
                  break;
               case DOTTED_SMALL_THIN:
                  stroke = dotSmallThinStroke;
                  break;
            }
            
            g2.setStroke(stroke);
            break;
         case SET_FONT:
            g2.setFont(ps.font.font);
            break;
         case SET_FONT_STYLE:
            g2.setFont(g2.getFont().deriveFont(convertFontStyle(ps.style)));
            break;
         case TRANSLATE:
            g2.translate(ps.x, ps.y);
            translateX += ps.x;
            translateY += ps.y;
            break;
         case CLEAR_AREA:
            if (ps.color != null)
            {
               g2.setColor(SwingHelper.fromColorRgb(ps.color));
            }
            g2.fillRect(ps.x, ps.y, ps.width, ps.height);
            break;
         case CLEAR:
            if (ps.color != null)
            {
               g2.setColor(SwingHelper.fromColorRgb(ps.color));
            }
            g2.fillRect(ps.x, ps.y, ps.width, ps.height);
            break;
         case CLIP:
            java.awt.Rectangle prevClip = g2.getClipBounds();
            if (prevClip != null)
            {
               clippings.push(prevClip);
            }
            
            // debugging aid: enable these lines to outline the clipped area
            if (false)
            {
               g2.setColor(Color.green);
               g2.drawRect(ps.x, ps.y, ps.width, ps.height);
            }
            
            g2.clipRect(ps.x, ps.y, ps.width, ps.height);
            break;
         case NO_CLIP:
            java.awt.Rectangle clip = clippings.isEmpty() ? null : clippings.pop();
            g2.setClip(clip);
            break;
         case SET_LOCATION:
            frame.setLocation(ps.x, ps.y);
            break;
         case RESIZE_WINDOW:
            Dimension sz = new Dimension(ps.width, ps.height);

            pane.setPreferredSize(sz);
            pane.setMinimumSize(sz);
            pane.setMaximumSize(sz);
            
            frame.setSize(sz);
            pane.setSize(sz);
            
            frame.pack();

            resizeScreenBuffers(ps.width, ps.height);
            break;
         case DRAW_HIGHLIGHT:
            drawHighlightWorker(ps.x, ps.y, ps.width, ps.height, ps.color);
            break;
         case CLEAR_HIGHLIGHT:
            clearHighlightWorker();
            break;
         case SET_TITLE:
            frame.setTitle(ps.title);
         default:
            break;
      }      
   }
   
   /**
    * Obtain the content pane for the terminal simulator.
    */
   public ContentPane getContentPane()
   {
      return pane;
   }
   
   /**
    * Returns display size.
    * 
    * @return   The dimensions in pixels.
    */
   @Override
   public NativeDimension getDisplaySize()
   {
      Rectangle displayBounds = frame.getGraphicsConfiguration().getBounds(); 
      return new NativeDimension(displayBounds.width, displayBounds.height);
   }

   /**
    * Returns an application-usable display area. For example, a task bar is
    * not counted as the usable area.
    * 
    * @return   The area in pixels.
    */
   @Override
   public NativeRectangle getDisplayWorkArea()
   {
      Insets is = frame.getToolkit().getScreenInsets(frame.getGraphicsConfiguration());
      return new NativeRectangle(is.top, is.left, is.bottom, is.right);
   }
   
   /**
    * Render the given paint operation.  Any graphics context or other resources needed must
    * have already been setup/stored by the caller.
    */
   @Override
   public void flush()
   {                                                     
      graphics.drawImage(screenBuffer, 0, 0, pane);
   }
   
   /**
    * Obtain the lock object that can be used to synchronize drawing operations.
    *
    * @return   The shared lock object.
    */
   @Override
   protected Object lock()
   {
      return pane.getTreeLock();
   }
   
   /**
    * Get the metrics of the given font.  If the {@link #g2 graphics} have been obtained, call the
    * {@link #getFontMetrics} and the graphics object.  Otherwise, use the simulator's graphics.
    *  
    * @param    font
    *           The font for which the font metrics are to be obtained.
    *           
    * @return   The font metrics.
    */
   private FontMetrics getNativeFontMetrics(Font font)
   {
      synchronized (lock())
      {
         return (g2 == null) ? pane.getFontMetrics(font) : g2.getFontMetrics(font);
      }
   }

   /**
    * Resize the screen buffers so that they will have the new width/height.
    * <p>
    * This will be a no-op if either the width or the height is not a positive number.
    * 
    * @param   width
    *          The new width.
    * @param   height
    *          The new height.
    */
   private void resizeScreenBuffers(int width, int height)
   {
      if (width <= 0 || height <= 0)
      {
         return;
      }
      
      synchronized (lock())
      {
         BufferedImage resizedImage = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
         
         if (screenBuffer != null)
         {
            Graphics2D g = resizedImage.createGraphics();
            g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
            g.drawImage(screenBuffer, 0, 0, null);
            g.dispose();
         }
         
         screenBuffer = resizedImage;
         
         if (g2 != null)
         {
            g2.dispose();
         }
   
         g2 = screenBuffer.createGraphics();
         g2.translate(translateX, translateY);
         g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF);
         
         if (!clippings.isEmpty())
         {
            g2.clip(clippings.peek());
         }
   
         lock().notifyAll();
         repaint();
      }
   }

   /**
    * Convert the generic font style to a Swing font style.
    *
    * @param    style
    *           The font style to convert.
    *
    * @return   The Java Swing font style.
    */
   private int convertFontStyle(FontStyle style)
   {
      int result = 0;
      
      switch (style)
      {
         case BOLD:
            result = Font.BOLD;
            break;
         case ITALIC:
            result = Font.ITALIC;
            break;
         case BOLD_ITALIC:
            result = Font.BOLD | Font.ITALIC;
            break;
         case PLAIN:
            result = Font.PLAIN;
            break;
      }
      
      return result;
   }
   
   /**
    * Converts the image to one having transparent color.
    * 
    * @param   img
    *          Initial image to convert.
    * @param   transPix
    *          The color to be considered as transparent.
    * 
    * @return   The image with one tarnsparent color.
    */
   private BufferedImage getTransparentImage(BufferedImage img, int transPix)
   {
      // define custom image producer
      ImageProducer ip = new FilteredImageSource(img.getSource(),
                                                 new TransparencyFilter(transPix));
      // convert image
      return getBufferedImage(Toolkit.getDefaultToolkit().createImage(ip));
   }
   
   /**
    * Converts the image replacing standard 3D colors with currently redefined.
    * 
    * @param   img
    *          Initial image to convert.
    * 
    * @return   The image with replaced 3D colors color.
    */
   private BufferedImage get3DColorsConvertedImage(BufferedImage img)
   {
      // define custom image producer
      ImageProducer ip = new FilteredImageSource(img.getSource(),
                            new Convert3DColorsFilter(SystemColor.controlLtHighlight.getRGB(),
                                                      SystemColor.control.getRGB(),
                                                      SystemColor.controlShadow.getRGB(),
                                                      SystemColor.controlText.getRGB()));
      // convert image
      return getBufferedImage(Toolkit.getDefaultToolkit().createImage(ip));
   }
   
   /**
    * Converts the image into buffered image.
    * 
    * @param   imgToConvert
    *          Initial image to convert.
    * 
    * @return   The buffered image for further use.
    */
   private BufferedImage getBufferedImage(java.awt.Image imgToConvert)
   {
      // create empty image
      BufferedImage imgRes = new BufferedImage(imgToConvert.getWidth(null),
                                               imgToConvert.getHeight(null),
                                               BufferedImage.TYPE_INT_ARGB);

      // copy data to newly created image
      Graphics2D g2img = imgRes.createGraphics();
      g2img.drawImage(imgToConvert, 0, 0, null);
      g2img.dispose();
      
      return imgRes;
   }
    
   /**
    * Draw a string using the current font, which will be scaled so that the final string will
    * have the specified legacy width and height.
    * <p>
    * Once this API is finished, the graphic's font is restored to its state as before the call.
    *  
    * @param    ps
    *           The structure with the drawing details.
    */
   private void drawStringScaled(PaintStructure<Font, BufferedImage> ps)
   {
      Font origFont = g2.getFont();
      
      FontMetrics fm = getNativeFontMetrics(origFont);
      
      int txtWidth = fm.stringWidth(ps.text);
      LineMetrics lm = fm.getLineMetrics(ps.text, g2);
      int txtHeight = (int) lm.getHeight();
      
      if (txtWidth == ps.width && txtHeight == ps.height)
      {
         drawString(ps);
         return;
      }

      double scaleX = (double) ps.width / (double) txtWidth;
      double scaleY = (double) ps.height / (double) txtHeight;

      Font scaled = origFont.deriveFont(AffineTransform.getScaleInstance(scaleX, scaleY));
      g2.setFont(scaled);
      try
      {
         fm = getNativeFontMetrics(scaled);
         lm = fm.getLineMetrics(ps.text, g2);

         drawString(ps);
      }
      finally
      {
         g2.setFont(origFont);
      }
   }

   /**
    * Draw a string using the current font.
    *  
    * @param    ps
    *           The structure with the drawing details.
    */                                               
   private void drawString(PaintStructure<Font, BufferedImage> ps)
   {
      int y = ps.y;
      if (ps.centered)
      {
         FontMetrics fm = g2.getFontMetrics();
         LineMetrics lm = fm.getLineMetrics(ps.text, g2);

         y = (ps.y - fm.getHeight()) / 2 + Math.round(lm.getAscent());
      }

      g2.drawString(ps.text, ps.x, y);
   }

   /**
    * This is the worker method for drawing a highlighted rectangular area of the specified
    * color and bounds. The color is alpha-blended with the content of the screen buffer.
    * 
    * @param    x
    *           left position
    * @param    y
    *           top position
    * @param    width
    *           area width
    * @param    height
    *           area height
    * @param    color
    *           the highlight color
    */
   private void drawHighlightWorker(int x, int y, int width, int height, ColorRgb color)
   {
      // keep the original composite and color to restore later
      Composite prevComposite = g2.getComposite();
      Color     prevColor     = g2.getColor();
      Color     newColor      = SwingHelper.fromColorRgb(color);
      
      try 
      {
         // save the area for later to allow clear
         highlightRestoreBuffer = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
         Graphics2D gh = highlightRestoreBuffer.createGraphics();
         int sx = x - translateX;
         int sy = y - translateY;
         gh.drawImage(screenBuffer, 0, 0, width, height, sx, sy, sx+width, sy+height, pane);
         
         // create and set new composite for the highlight
         AlphaComposite ac = AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.5f);
         g2.setComposite(ac);
   
         // draw the highlight
         g2.setColor(newColor);
         g2.fillRect(x - translateX, y - translateY, width, height);
         highlightX = x;
         highlightY = y;
      }
      finally
      {
         g2.setColor(prevColor);;
         g2.setComposite(prevComposite);
      }
   }
   
   /**
    * This is the worker method for restoring the screen area previously highlighted by
    * {@link #drawHighlightWorker(int, int, int, int, ColorRgb)}.
    */
   private void clearHighlightWorker()
   {
      if (highlightRestoreBuffer == null)
      {
         // nothing to restore
         return;
      }
      
      int w = highlightRestoreBuffer.getWidth();
      int h = highlightRestoreBuffer.getHeight();
      
      int dx = highlightX - translateX;
      int dy = highlightY - translateY;
      
      Composite prevComposite = g2.getComposite();
      AlphaComposite ac = AlphaComposite.getInstance(AlphaComposite.SRC, 1f);
      g2.setComposite(ac);
      g2.drawImage(highlightRestoreBuffer, dx, dy, dx+w, dy+h, 0, 0, w, h, pane);
      g2.setComposite(prevComposite);
      
      highlightRestoreBuffer = null;
   }
   
   /**
    * Custom RGB filter to change image transparency data on the fly.
    */
   static class TransparencyFilter
   extends RGBImageFilter
   {
      /** Alpha channel constant. */
      private static final int TRANSPARENCY_MASK = 0xFF000000;
      
      /** Pixel to use as transparent marker. */
      int rgbCheck;
      
      /**
       * Creates new filter with the given transparent pixel value.
       *
       * @param   pixTrans
       *          The pixel sample we will consider as transparent.
       */
      TransparencyFilter(int pixTrans)
      {
         rgbCheck = pixTrans | TRANSPARENCY_MASK;
      }

      /**
       * This is on the fly converter to turn on transparency for every pixels the sames as
       * one we defined as marker on the filter construction.  This method is calling when image
       * drawing is happening.
       *
       * @param   x
       *          The X pixel coordinate.
       * @param   y
       *          The Y pixel coordinate.
       * @param   pixel
       *          The ARGB pixel value of the original image.
       * 
       * @return   The new ARGB pixel value after transformation to render.
       */
      public int filterRGB(int x, int y, int pixel)
      {
         return (pixel | TRANSPARENCY_MASK) == rgbCheck ? pixel & ~TRANSPARENCY_MASK : pixel; 
      }
   }

   /**
    * Custom RGB filter to change image 3D colors on the fly.
    */
   static class Convert3DColorsFilter
   extends RGBImageFilter
   {
      /** Original 3D white color(255, 255, 255). */
      private static final int COLOR_3D_WHITE   = 0x00FFFFFF;
      
      /** Original 3D light gray color(192, 192, 192). */
      private static final int COLOR_3D_LT_GRAY = 0x00C0C0C0;
      
      /** Original 3D dark gray color(128, 128, 128). */
      private static final int COLOR_3D_DK_GRAY = 0x00808080;
      
      /** Original 3D black color(0, 0, 0). */
      private static final int COLOR_3D_BLACK   = 0x00000000;
      
      /** New value for 3D white. */
      int newWhite;

      /** New value for 3D light gray. */
      int newLtGray;
      
      /** New value for 3D dark gray. */
      int newDkGray;
      
      /** New value for 3D black. */
      int newBlack;
      
      /**
       * Creates new filter with the given 3D colors to replace.
       *
       * @param   white
       *          The new value for white color.
       * @param   ltGray
       *          The new value for light gray color.
       * @param   dkGray
       *          The new value for dark gray color.
       * @param   black
       *          The new value for dark gray color.
       */
      Convert3DColorsFilter(int white, int ltGray, int dkGray, int black)
      {
         newWhite  = white;
         newLtGray = ltGray;
         newDkGray = dkGray;
         newBlack  = black;
      }

      /**
       * This is on the fly converter to change the 3D colors with one we passed on the filter
       * construction.  This method is calling when image drawing is happening.
       *
       * @param   x
       *          The X pixel coordinate.
       * @param   y
       *          The Y pixel coordinate.
       * @param   pixel
       *          The ARGB pixel value of the original image.
       * 
       * @return   The new ARGB pixel value after transformation to render.
       */
      public int filterRGB(int x, int y, int pixel)
      {
         return (pixel | COLOR_3D_WHITE) == COLOR_3D_WHITE ? newWhite : 
                (pixel | COLOR_3D_LT_GRAY) == COLOR_3D_LT_GRAY ? newLtGray :
                (pixel | COLOR_3D_DK_GRAY) == COLOR_3D_DK_GRAY ? newDkGray :
                (pixel | COLOR_3D_BLACK) == COLOR_3D_BLACK ? newBlack : pixel; 
      }
   }
}
